/*
  Copyright 2003 Rikard Bjrklind, Mikael Gransell

  This file is part of dc-qt.

  dc-qt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  dc-qt is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with dc-qt; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <qlayout.h>
#include <qvbox.h>
#include <qhbox.h>
#include <qtabwidget.h>
#include <qpushbutton.h>
#include <qmessagebox.h>
#include <qinputdialog.h>
#include <qlabel.h>
#include <qfile.h>
#include <qdir.h>
#include <zlib.h>
#include "hubwdgt.h"
#include "debugdlg.h"
#include "searchdlg.h"
#include "dctc_hub.h"
#include "connect_dlg.h"
#include "dc_hub.h"
#include "dc_settings.h"
#include "util.h"
#include "transfer_widget.h"
#include "chatwdgt.h"

HubWdgt::HubWdgt( transfer_widget* transferWin, QWidget* parent, const char* name ) : QWidget( parent, name )
{
     this->transferWin = transferWin;

     QGridLayout* layout = new QGridLayout( this, 4, 1 );

     // Tab bar for the chat and user tabs
     tabs = new QTabWidget( this, "Tab bar" );

     // Public chat
     publicChat = new ChatWdgt(false,this,"public chat");
     
     // User list
     users = new UserList( tabs, "User list" );
    
     tabs->addTab( publicChat, "Ch&at" );
     tabs->addTab( users, "U&sers" );
     
     // Add all widgets to our layout
     layout->addWidget( tabs, 0, 0 );
     
     hubInterface = NULL;
     hubs = new dc_hub();
     name = "New Hub";
     searchDlg = 0;
}

bool HubWdgt::connect()
{
     // Bring up the connection dialog
     ConnectDlg *dlg = new ConnectDlg(hubs, this, "connect_dlg", true);
     if (dlg->exec() == QDialog::Accepted)
     {
	  disconnect();                      // Disconnect if already connected somewhere.
	  settings.save_favourites_file();   // To save the new favourites if any 
	  hub_info hub = dlg->selected_hub();
	  hubInterface = new dctc_hub(0, "hub-" + hub.ip);
      
	  if (hub.ip == QString::null)
	       return false;
      
	  // if we connected to an already running hub, send the pid
	  // to the connect function as well
	  int ret;
	  if (dlg->selected_running())
	       ret = hubInterface->hubconnect(hub.ip, hub.pwd,hub.pid);
	  else
	       ret = hubInterface->hubconnect(hub.ip,hub.pwd);
      
	  if (ret != CONNECT_SUCCESS)
	  {
	       QMessageBox::critical(this, "Connect", "Could not connect to hub");
	       delete hubInterface;
	       hubInterface = 0;
	  }
     }
     
     // TODO Adda massa connects hr
     if( hubInterface ) 
     {
       // File transfers

       // Uploads
       QObject::connect( hubInterface, SIGNAL(uploadAdded(const QString&)),
			 transferWin, SLOT(new_ul_info(const QString&)));
       QObject::connect( hubInterface, SIGNAL(uploadUpdated(const QString&)),
			 transferWin, SLOT(update_ul_info(const QString&)));
       QObject::connect( hubInterface, SIGNAL(uploadClosed(const QString&, const QString&)),
			 transferWin, SLOT(closed_ul_info(const QString&, const QString&)));
       
       // GDLs
       QObject::connect( transferWin, SIGNAL(request_gdl_update()),
			 hubInterface, SLOT(request_dlist()));
       QObject::connect( hubInterface, SIGNAL(gdlUpdated(const QString&)),
			 transferWin, SLOT(update_gdl_info(const QString&)));
       QObject::connect( hubInterface, SIGNAL(gdlListEnd()),
			 transferWin, SLOT(update_gdl_end()));

       QObject::connect( transferWin, SIGNAL(start_gdl(int, const QString &, const QString &, const QString &, unsigned long int)),
			 hubInterface, SLOT(get_file(int, const QString &, const QString &, const QString &, unsigned long int)));

       QObject::connect( transferWin, SIGNAL(add_gdl_source(int, const QString &, const QString &, unsigned long int)),
			 hubInterface, SLOT(addSource(int, const QString &, const QString &, unsigned long int)));


       QObject::connect( transferWin, SIGNAL(close_gdl(unsigned int)),
			 hubInterface,SLOT(cancel_transfer(unsigned int)));
       QObject::connect(transferWin, SIGNAL(close_dl(unsigned int,const QString&,const QString&)),
			hubInterface,SLOT(delete_gdl(unsigned int,const QString&,const QString&)));

       // User file list
       QObject::connect( transferWin, SIGNAL(getUserFileList(const QString &)),
			 hubInterface, SLOT(get_user_list(const QString &)));

/*
	  // remove
	  QObject::connect( hubInterface, SIGNAL(new_transfer(int, const QString&)),
			    transferWin, SIGNAL(addTransfer(int, const QString&)));
	  QObject::connect( hubInterface, SIGNAL(transfer_info(int,const QString&,int,int,int,const QString&,const QString&)),
			    transferWin, SLOT(handleInfo(int,const QString&,int,int,int,const QString&,const QString&)));
	  // remove
	  QObject::connect( hubInterface, SIGNAL( active_transfers(  list<int>& )),
			    transferWin, SIGNAL(activeTransfers(  list<int>& )));
	  
	  QObject::connect( hubInterface, SIGNAL( new_upload(int,const QString&, const QString&)),
			    transferWin, SIGNAL( addTransfer(int,const QString&, const QString&)));
	  QObject::connect( hubInterface, SIGNAL( transfer_info(int,int,int)),
			    transferWin, SIGNAL( transferInfo(int,int,int)));
	  
	  QObject::connect(transferWin, SIGNAL(updateDList()), hubInterface, SLOT( request_dlist() ));
	  
	  QObject::connect(transferWin, SIGNAL(cancelTransfer( int )), hubInterface, SLOT(cancel_transfer( int )));
	  QObject::connect(transferWin, SIGNAL(retryTransfer(const QString&)),
			   hubInterface, SLOT(retryTransfer(const QString&)));
	  QObject::connect(transferWin, SIGNAL(getUserFileList(const QString&)),
			   hubInterface,SLOT(get_user_list(const QString&)));
	  QObject::connect(hubInterface, SIGNAL(done_transfer(const QString &, const QString &)), 
			   transferWin, SIGNAL(removeTransfer(const QString &, const QString&)));
	  QObject::connect(transferWin,SIGNAL(getFile(int,const QString&, const QString&, const QString&,int)),
			   hubInterface,SLOT(get_file(int,const QString&, const QString&, const QString&,int)));
	  QObject::connect(transferWin,SIGNAL(addSource(int,const QString&,const QString&,int)),
			   hubInterface,SLOT(addSource(int,const QString&,const QString&,int)));

	  QObject::connect(hubInterface, SIGNAL(gdlListEnd()), transferWin, SIGNAL(removeInactive()));
*/

	  QObject::connect(hubInterface, SIGNAL(unattachedGDL(const QString&)),
			   this, SLOT(queryUnattached(const QString&)));
	  
	  // Users
	  QObject::connect( hubInterface, SIGNAL(user_info(const QString&, const QString&, const QString&, const QString&)),
			    users,SLOT(addUser(const QString&, const QString&, const QString&, const QString&)));
	  QObject::connect( hubInterface, SIGNAL(user_added(const QString&, bool)), users,SLOT(addUser(const QString&, bool)));
	  QObject::connect( hubInterface, SIGNAL(user_removed(const QString&)), users, SLOT(removeUser(const QString&)));
	  QObject::connect( users, SIGNAL(getUserFiles( const QString& )),
			    hubInterface, SLOT(get_user_list( const QString& )));
	  QObject::connect( users, SIGNAL(getUserInfo(const QString&)), hubInterface, SLOT(get_uinfo(const QString&)));

	  // Public chat stuff
	  QObject::connect( publicChat, SIGNAL(newChatStr(const QString&,QWidget*)), 
			    this, SLOT(sendChatMes(const QString&,QWidget*)));
	  QObject::connect( hubInterface, SIGNAL(chat_msg(const QString&)), 
			    publicChat, SLOT( addChatStr(const QString&) ));
	  // Private chat stuff
	  QObject::connect( hubInterface, SIGNAL(privchat_msg(const QString&, const QString&)), 
			    this, SLOT(privChatMsg(const QString&, const QString&)) );
	  QObject::connect( users, SIGNAL(activateChat(const QString&)), 
			    this, SLOT( activateChat(const QString&)));

	  // Hubname
	  QObject::connect( hubInterface, SIGNAL( hub_name(const QString&)), this, SLOT(setHubName(const QString&)));

	  //QObject::connect( hubInterface, SIGNAL( user_flist_file( const QString&, const QString&, const QString& )),
	  //	    this, SLOT( addUserFile( const QString&, const QString&, const QString& )));
	  
	  QObject::connect( hubInterface, SIGNAL(userFileList(const QString&)), 
			    this, SLOT(loadUserList(const QString&)));
	  
	  QObject::connect(hubInterface, SIGNAL(passwordRequired()),
			   this, SLOT(requestPassword()));

	  QObject::connect(hubInterface, SIGNAL(hub_died()),
			   this, SLOT(disconnect()));
	  

	  hubInterface->get_ulist();

     }

     delete dlg;      
     return hubInterface!=0;
}

void HubWdgt::requestPassword()
{
    bool ok;
    QString password = QInputDialog::getText("Password required","Enter password:",QLineEdit::Password,
					     QString::null,&ok,this,"passworddialog");
    if(ok && !password.isEmpty())
	hubInterface->sendPassword( password );
}

void HubWdgt::setHubName(const QString& aName)
{
     name = aName;
     emit gotName(name,this);
}

/// Called when somone pressed 'send message' on the user list popup menu
void HubWdgt::activateChat(const QString& user)
{
     bool found = false;
     // Try to find an already existing tab
     for(int i=2; i < tabs->count(); i++)
     {
	  if( tabs->label(i) == user )
	  {
	       tabs->setCurrentPage(i);
	       found = true;
	  }
     }
     
     if(!found)    // Otherwise create a new tab
     {
	  ChatWdgt *pchat = new ChatWdgt(true, this, "private chat");
	  tabs->addTab( pchat, user );
	  QObject::connect( pchat, SIGNAL(newChatStr(const QString&, QWidget*)), 
			    this, SLOT(sendChatMes(const QString&, QWidget*)));
	  QObject::connect( pchat, SIGNAL(close(QWidget*)), tabs, SLOT(removePage(QWidget*)));
	  pchat->show();
     }
}

/// Called when a private chat message is received from the hub interface.
void HubWdgt::privChatMsg(const QString& user, const QString& msg)
{
    // The message could be from a user on another hub.
    // The message is on the form "<user (hubname)> message" if it comes from another hub.
    // It could also miss this tag in case of a hub message, which makes it impossible to
    // know which hub it originated from (bad dctc).
    if( msg[0]=='<' )
    {
	int i1 = msg.find('(');
	int i2 = msg.find(')');
	if(i1!=-1 && i2!=-1) {
	    QString hubName = msg.mid(i1+1,i2-i1-1);
	    debugWin->print("hubname="+hubName+"\n");
	    if(hubName != name) return;
	}
    }

     bool found = false;
     // Try to find the user among the tabs
     for(int i=2;i < tabs->count();i++)
     {
	  if(tabs->label(i) == user) 
	  {
	       ChatWdgt *pchat = (ChatWdgt *)tabs->page(i);
	       pchat->addChatStr(msg);
	       found = true;
	       break;
	  }
     }

     if(!found)        // The user was not found, add a new tab
     {
	  ChatWdgt *pchat = new ChatWdgt(true, this, "private chat");
	  tabs->addTab( pchat, user );
	  pchat->addChatStr(msg);
	  QObject::connect( pchat, SIGNAL(newChatStr(const QString&, QWidget*)), 
			    this, SLOT(sendChatMes(const QString&, QWidget*)));
	  QObject::connect( pchat, SIGNAL(close(QWidget*)), tabs, SLOT(removePage(QWidget*)));
     }
}

/// Called when a ChatWdgt has a new chatstring to send
void HubWdgt::sendChatMes(const QString& msg,QWidget* originator)
{
     // If the originator is the public chat just send it
     if( originator == tabs->page(0) )
	  hubInterface->send_chatmsg(msg);
     // Otherwise it is a private chat msg, so find the tab and the user
     else 
     {
	  for(int i=2;i < tabs->count();i++)
	       if( tabs->page(i) == originator ) {
		    hubInterface->send_privmsg( msg, tabs->label(i));
		    break;
	       }
     }

}

void HubWdgt::disconnect()
{
    debugWin->print("Disconnecting...\n");
     if( hubInterface ) {
	 hubInterface->hubdisconnect();
	 delete hubInterface;
	 hubInterface = NULL;
	 emit disconnected(id);
     }
}

void HubWdgt::openSearchDlg()
{
  if( hubInterface )
    {
      if( searchDlg ) searchDlg->show();
      else 
	{
	  // Create new search dialog
	  //searchDlg = new SearchDlg( this, "Search dialog" );
	  searchDlg = new SearchDlg();
	  searchDlg->setCaption("Search@"+name);
	  searchDlg->show();
	  QObject::connect( searchDlg, SIGNAL(search(const QString&, bool)), 
			    hubInterface, SLOT(search_for( const QString&, bool)) );
	  
	  QObject::connect( hubInterface, 
			    SIGNAL(search_file(const QString&, const QString&, const QString &, const QString &,const QString&)),
			    this, SLOT(handleSearchResult(const QString&,const QString&, const QString&, const QString &, const QString &)));
	  
	  QObject::connect( searchDlg, SIGNAL( downloadFile(const QString&, const QString&, const QString&,unsigned long int)),
			    transferWin, SLOT( download_file(const QString&, const QString&, const QString&,unsigned long int)));
	  /*
	    QObject::connect( searchDlg, SIGNAL( downloadFiles(list<FileInfo>)),
	    transferWin, SLOT( downloadFiles(list<FileInfo>)));
	  */
	  QObject::connect( searchDlg, SIGNAL( getUserFiles( const QString& )),
			    hubInterface, SLOT(get_user_list( const QString& )));
	} 
    }
}

void HubWdgt::handleSearchResult(const QString &user,
				 const QString &file,
				 const QString &size,
				 const QString &slot,
				 const QString &hubname)
{
    if(hubname == name) // This result is for us
    {
	openSearchDlg();
	searchDlg->addSearchResult(user,file,size,slot);
    }
    else
    {
	// If the last search was a multihub one, inform the others
	if(searchDlg->isMultiSearch())
	{
	    emit multiSearchResult(user,file,size,slot,hubname);
	}
    }
}

void HubWdgt::removeFileDlg( const QString& user )
{
     list<UserFileDlg *>::iterator it;
     for(it = dialogs.begin(); it != dialogs.end(); it++)
     {
	  UserFileDlg* tmp = *it;
	  if( tmp->getUserName() == user )
	  {
	       dialogs.remove( tmp );
	       break;
	  }
     }   
}

#define LS_CACHE_MAGIC	0x4C536430

struct LsCacheHeader
{
     unsigned int ls_magic;
     unsigned int packed_size;
     unsigned int unpacked_size;
     unsigned int pack_algorithm;
     time_t retrieve_time;
     unsigned long long share_size;
};

#define LS_CACHE_PACK_NONE  0
#define LS_CACHE_PACK_BZIP  1

void HubWdgt::loadUserList( const QString& user )
{
  	QString filename = QString(getenv("HOME")) + "/.dctc/ls_cache/" + user;   
	char *ret = NULL;

	FILE *f = fopen(filename,"rb");
	if(f!=NULL)
	{
	     LsCacheHeader header;
	     if(fread(&header,1,sizeof(LsCacheHeader),f) != sizeof(LsCacheHeader))
	     {
		  debugWin->print("Error reading ls_cache header.\n");
		  fclose(f);
		  return;
	     }

	     if( header.ls_magic != LS_CACHE_MAGIC )
	     {	
		  debugWin->print("Invalid ls_cache header.\n");
		  fclose(f);
		  return;
	     }
	     
	     switch(header.pack_algorithm)
	     {
	     case LS_CACHE_PACK_NONE:		/* result no compressed */
		  ret = (char *)malloc(header.unpacked_size+1);
		  if(ret==NULL)
		  {
		       fclose(f);
		       return;
		  }
		  
		  if(fread( ret,1,header.unpacked_size,f) != header.unpacked_size )
		  {
		       debugWin->print("Trucated ls_cache file detected.\n");
		       fclose(f);
		       free(ret);
		       return;
		  }
		  fclose(f);
		  ret[header.unpacked_size]='\0';
		  break;
		  
	     case LS_CACHE_PACK_BZIP:
	     {
		  char *pack_area;
		  unsigned long upack;
		  pack_area=(char*)malloc(header.packed_size);
		  if( pack_area==NULL )
		  {
		       fclose(f);
		       return;
		  }
		  if( fread(pack_area,1,header.packed_size,f) != header.packed_size )
		  {
		       debugWin->print("Trucated ls_cache file detected.\n");
		       fclose(f);
		       free(pack_area);
		       return;
		  }
		  fclose(f);
		  
		  ret=(char*)malloc(header.unpacked_size+1);
		  if( ret==NULL )
		  {
		       free(pack_area);
		       return;
		  }
		  
		  upack = header.unpacked_size;
		  if( uncompress((Bytef*)ret,&upack,(Bytef*)pack_area,header.packed_size) != Z_OK )
		  {
		       debugWin->print("Failed to uncompress ls_cache file.\n");
		       free(pack_area);
		       free(ret);
		       return;
		  }
		  
		  free(pack_area);
		  if( upack != header.unpacked_size )
		  {
		       debugWin->print("Invalid decompressed result.\n");
		       free(ret);
		       return;
		  }
		  ret[header.unpacked_size]='\0';
	     }
	     break;
	     
	     default:
		  debugWin->print("loadUserList: unknown compression algorithm.\n");
		  fclose(f);
		  return;
	     }
	}
	else
	{
	     debugWin->print("Unable to open ls_cache file: " + filename + "\n");
	}

	UserFileDlg* temp = new UserFileDlg( user, this, "user file dlg" );
	temp->show();
	dialogs.push_back( temp );
	QObject::connect( temp, SIGNAL(killed( const QString& )),
			  this, SLOT(removeFileDlg( const QString& )));

	// estrato replacing this:
	//	QObject::connect( temp, SIGNAL( downloadFile(const QString&, const QString&,const QString&, int)), 
	//			  hubInterface, SLOT(get_file(const QString&, const QString&,const QString&, int)));
	//
	// with this:
	QObject::connect( temp, SIGNAL( downloadFile(const QString&, const QString&, const QString&,unsigned long int)),
			  transferWin, SLOT( download_file(const QString&, const QString&, const QString&,unsigned long int)));

	QStringList strLst = QStringList::split(QChar('\n'), QString(ret));
	for ( QStringList::Iterator it = strLst.begin(); it != strLst.end(); ++it ) 
	{
	     temp->addFile( (*it).section('|',0,0), 
			    (*it).section('|',1,1) );
	}
}


/// This slot receives the signal unattachedGDL from the hub interface and asks the user what to do.
void HubWdgt::queryUnattached(const QString& name)
{
    QDir dir(settings.get_setting(dc_settings::DLDIR) + "GDL/" + name,QString::null,QDir::Name|QDir::IgnoreCase,
	     QDir::Files|QDir::Hidden);
    QStringList files = dir.entryList();

    switch( QMessageBox::information(this,
				     "File exists",
				     "The following file is partly downloaded:\n"+name+"What do you want to do?",
				     "&Resume", "Re&start", "&Cancel",0,2) )
    {
    case 0:
	// Resume
	hubInterface->attachFile(name);
	break;
    case 1:
	// Restart

	for(QStringList::iterator i=files.begin();i!=files.end();i++)
	{
	    if(*i=="." || *i=="..") continue;
	    debugWin->print("Removing file: " + dir.absPath()+"/"+*i+"\n");
	    if(!QFile::remove(dir.absPath()+"/"+*i))
	    {
		QMessageBox::critical(this,"dc-qt: Error", 
				      "Error removing file:\n" + *i);
		break;
	    }
	}
	if(!dir.rmdir(dir.absPath()))
	{
	    QMessageBox::critical(this,"dc-qt: Error", 
				  "Error removing directory:\n" + dir.absPath());
	    break;
	}
	
//	hubInterface->restart();

//	estrato TODO: transferWin->restart( name );

	break;
    case 2:
	// Cancel, do nothing.
	break;
    default:
	;
    }
}
